@@ -45,7 +45,7 @@ class SmartDriver(SeleniumDriverCore):
         msg = ''
         if self.test_case_creation_mode:
             self._test_case_upload_screenshot(element_name)
-            element_box = self._test_case_get_box(element_name)
+            element_box, needs_reload = self._test_case_get_box(element_name)
             if element_box:
                 real_elem = self._match_bounding_box_to_web_element(element_box, multiplier=self.multiplier)
                 element = ai_elem(real_elem.parent, real_elem, element_box, self.driver,
@@ -57,7 +57,7 @@ class SmartDriver(SeleniumDriverCore):
                                                                                                        label_url))
                 webbrowser.open(label_url)
                 while True:
-                    element_box = self._test_case_get_box(element_name)
+                    element_box, needs_reload = self._test_case_get_box(element_name)
                     if element_box is not None:
                         print('Element was labeled, moving on')
                         real_elem = self._match_bounding_box_to_web_element(element_box, multiplier=self.multiplier)
@@ -92,7 +92,12 @@ class SmartDriver(SeleniumDriverCore):
                 }
                 classify_url = self.url + '/detect'
                 start = time.time()
-                r = requests.post(classify_url, json=data, verify=False)
+                for _ in range(2):
+                    try:
+                        r = requests.post(classify_url, json=data, verify=False, timeout=10)
+                        break
+                    except Exception:
+                        log.error('error during detect')
                 end = time.time()
                 if self.debug:
                     print(f'Classify time: {end - start}')
