@@ -2,6 +2,7 @@ import base64
 import logging
 import requests
 import time
+import uuid
 import urllib.parse
 import warnings
 import webbrowser
@@ -11,6 +12,9 @@ import io
 from distutils.util import strtobool
 from PIL import Image
 from packaging import version
+from time import sleep
+
+
 import selenium
 
 if version.parse(selenium.__version__) < version.parse('4.0.0'):
@@ -25,20 +29,23 @@ from selenium.common.exceptions import StaleElementReferenceException
 from selenium import webdriver
 
 from .utils.selenium_core import SeleniumDriverCore
+
 requests.packages.urllib3.disable_warnings()
 
 
 log = logging.getLogger(__name__)
 
 class SmartDriver(SeleniumDriverCore):
-    def __init__(self, driver, api_key, initialization_dict={}):
-        self.version = 'selenium-0.1.20'
+    def __init__(self, driver, api_key=None, initialization_dict={}):
+        self.version = 'selenium-0.1.14'
         SeleniumDriverCore.__init__(self, driver, api_key, initialization_dict)
         window_size = self.driver.get_window_size()
+        self.window_size = window_size
         screenshotBase64 = self._get_screenshot()
 
         im = Image.open(io.BytesIO(base64.b64decode(screenshotBase64)))
         width, height = im.size
+        self.im_size = im.size
         self.multiplier = 1.0 * width / window_size['width']
         self_attrs = dir(self)
         # Disable warnings
@@ -52,11 +59,28 @@ class SmartDriver(SeleniumDriverCore):
             screenshotBase64 = self.driver.get_screenshot_as_base64()
         return screenshotBase64
 
+    def process_exist_screenshot_response(self, element_name, key, msg, tresp_data):
+        if self.debug:
+            print(f'Found cached box in action info for {element_name} using that')
+        element_box = tresp_data['predicted_element']
+        expected_offset = tresp_data['page_offset']
+        element_box['y'] += expected_offset
+
+        if self.use_cdp:
+            parent_elem = None
+            real_elem = element_box
+        else:
+            real_elem = self._match_bounding_box_to_selenium_element(element_box, multiplier=self.multiplier)
+            parent_elem = real_elem.parent
+        element = ai_elem(parent_elem, real_elem, element_box, self.driver,
+                          self.multiplier)
+        return element, key, msg
+
     def _classify(self, element_name):
         msg = ''
         if self.test_case_creation_mode:
             self._test_case_upload_screenshot(element_name)
-            element_box = self._test_case_get_box(element_name)
+            element_box, needs_reload = self._test_case_get_box(element_name)
             if element_box:
                 if self.use_cdp:
                     parent_elem = None
@@ -67,12 +91,13 @@ class SmartDriver(SeleniumDriverCore):
                 element = ai_elem(parent_elem, real_elem, element_box, self.driver, self.multiplier)
                 return element, self.last_test_case_screenshot_uuid, msg
             else:
-                label_url = self.url + '/testcase/label?test_case_name=' + urllib.parse.quote(self.test_case_uuid)
+                event_id = str(uuid.uuid4())
+                label_url = f'{self.url}/testcase/label?test_case_name={urllib.parse.quote(self.test_case_uuid)}&event_id={event_id}'
                 log.info('Waiting for bounding box of element {} to be drawn in the UI: \n\t{}'.format(element_name,
                                                                                                        label_url))
                 webbrowser.open(label_url)
                 while True:
-                    element_box = self._test_case_get_box(element_name)
+                    element_box, needs_reload = self._test_case_get_box(element_name, event_id=event_id)
                     if element_box is not None:
                         if self.use_cdp:
                             parent_elem = None
@@ -84,6 +109,10 @@ class SmartDriver(SeleniumDriverCore):
                         element = ai_elem(parent_elem, real_elem, element_box, self.driver,
                                           self.multiplier)
                         return element, self.last_test_case_screenshot_uuid, msg
+
+                    if needs_reload:
+                        self._test_case_upload_screenshot(element_name)
+
                     time.sleep(2)
         else:
             element = None
@@ -93,20 +122,25 @@ class SmartDriver(SeleniumDriverCore):
             screenshotBase64 = self._get_screenshot()
             key = self.get_screenshot_hash(screenshotBase64)
             resp_data = self._check_screenshot_exists(key, element_name)
+
             if resp_data['success'] and 'predicted_element' in resp_data and resp_data['predicted_element'] is not None:
                 log.debug(resp_data.get("message", ""))
-                if self.debug:
-                    print(f'Found cached box in action info for {element_name} using that')
-                element_box = resp_data['predicted_element']
-                if self.use_cdp:
-                    parent_elem = None
-                    real_elem = element_box
-                else:
-                    real_elem = self._match_bounding_box_to_selenium_element(element_box, multiplier=self.multiplier)
-                    parent_elem = real_elem.parent
-                element = ai_elem(parent_elem, real_elem, element_box, self.driver,
-                                  self.multiplier)
-                return element, key, msg
+                current_offset = self.driver.execute_script('return window.pageYOffset;')
+                bottom_page_offset = current_offset + self.window_size['height']
+
+                real_offset = resp_data['page_offset'] / self.multiplier
+                if real_offset > bottom_page_offset or real_offset < current_offset:
+                    scroll_offset = int(real_offset - current_offset)
+                    ActionChains(self.driver).scroll_by_amount(0, scroll_offset).perform()
+
+                    sleep(1)
+                    screenshotBase64 = self._get_screenshot()
+                    key = self.get_screenshot_hash(screenshotBase64)
+                    resp_data = self._check_screenshot_exists(key, element_name)
+                if resp_data['success'] and 'predicted_element' in resp_data and resp_data[
+                    'predicted_element'] is not None:
+                    return self.process_exist_screenshot_response(element_name, key, msg, resp_data)
+
             source = ''
 
             # Check results
@@ -118,22 +152,32 @@ class SmartDriver(SeleniumDriverCore):
                         'test_case_name': self.test_case_uuid}
                 classify_url = self.url + '/detect'
                 start = time.time()
-                r = requests.post(classify_url, json=data, verify=False)
+                for _ in range(2):
+                    try:
+                        r = requests.post(classify_url, json=data, verify=False, timeout=10)
+                        break
+                    except Exception:
+                        log.error('error during detect')
                 end = time.time()
                 if self.debug:
                     print(f'Classify time: {end - start}')
                 response = r.json()
                 if not response['success']:
-                    classification_error_msg = response['message'].replace(self.default_prod_url, self.url)
-                    log.debug(classification_error_msg)
-                    raise Exception(classification_error_msg)
-
+                    # Could be a failure or an element off screen
+                    response = self._classify_full_screen(element_name)
+                    if not response['success']:
+                        classification_error_msg = response['message'].replace(self.default_prod_url, self.url)
+                        log.debug(classification_error_msg)
+                        raise Exception(classification_error_msg)
                 run_key = response['screenshot_uuid']
                 msg = response.get('message', '')
                 msg = msg.replace(self.default_prod_url, self.url)
                 log.debug(msg)
 
                 element_box = response['predicted_element']
+                self.page_offset = self.driver.execute_script('return window.pageYOffset')
+                element_box['y'] += self.page_offset * self.multiplier
+
                 if self.use_cdp:
                     parent_elem = None
                     real_elem = element_box
@@ -141,10 +185,61 @@ class SmartDriver(SeleniumDriverCore):
                     real_elem = self._match_bounding_box_to_selenium_element(element_box, multiplier=self.multiplier)
                     parent_elem = real_elem.parent
                 element = ai_elem(parent_elem, real_elem, element_box, self.driver, self.multiplier)
-            except Exception:
+            except Exception as e:
+                log.error(e)
                 logging.exception('exception during classification')
             return element, run_key, msg
 
+    def _classify_full_screen(self, element_name):
+        ActionChains(self.driver).scroll_by_amount(0, -100000).perform() # go to the top
+        last_offset = -1
+        offset = 1
+        responses = []
+        window_height = self.window_size['height']
+        r = None
+        while offset > last_offset:
+            last_offset = offset
+            ActionChains(self.driver).scroll_by_amount(0, int(window_height)).perform()
+            sleep(0.2)
+            offset = self.driver.execute_script('return window.pageYOffset')
+            screenshotBase64 = self._get_screenshot()
+            data = {'screenshot': screenshotBase64,
+                    'source': '',
+                    'api_key': self.api_key,
+                    'label': element_name,
+                    'test_case_name': self.test_case_uuid}
+            classify_url = self.url + '/detect'
+            r = requests.post(classify_url, json=data, verify=False).json()
+            if r['success']:
+                return r
+        return r
+
+    def _pick_best_fs_response(self, responses):
+        """
+        First we check if we have zero success, if yes we pick the first one,
+        one unique success, if yes we pick that
+        more than that, we pick the one with highest score.
+        """
+
+        success_ct = len([r for r in responses if r['success']])
+        if success_ct == 0:
+            return responses[0]
+        elif success_ct == 1:
+            for r in responses:
+                if r['success']:
+                    return r
+        else:
+            best_score = 0
+            best_response = None
+            for r in responses:
+                if r['success']:
+                    if r['score'] > best_score:
+                        best_score = r['score']
+                        best_response = r
+            return best_response
+
+
+
     def _match_bounding_box_to_selenium_element(self, bounding_box, multiplier=1):
         """
             We have to ba hacky about this becasue Selenium does not let us click by coordinates.
@@ -154,7 +249,7 @@ class SmartDriver(SeleniumDriverCore):
         new_box = {'x': bounding_box['x'] / multiplier, 'y': bounding_box['y'] / multiplier,
                    'width': bounding_box['width'] / multiplier, 'height': bounding_box['height'] / multiplier}
 
-        element_types = ["img", "input", "button", "a", "*"]
+        element_types = ["a", "input", "button", "img",  "*"]
         for element_type in element_types:
             # Get all elements
             try:
@@ -261,8 +356,10 @@ class ai_elem(webdriver.remote.webelement.WebElement):
                 self.driver.execute_script('document.elementFromPoint(%d, %d).click();' % (int(self._cx), int(self._cy)))
         else:
             # Multiplier needs to be undone as js doesn't care about it. only selenium/appium
+            self.driver.execute_cdp_cmd('Input.dispatchMouseEvent', { 'type': 'mouseMoved', 'button': 'left', 'clickCount': 1, 'x': self._cx, 'y': self._cy})
+            time.sleep(0.07)
             self.driver.execute_cdp_cmd('Input.dispatchMouseEvent', { 'type': 'mousePressed', 'button': 'left', 'clickCount': 1, 'x': self._cx, 'y': self._cy})
-            time.sleep(0.05)
+            time.sleep(0.15)
             self.driver.execute_cdp_cmd('Input.dispatchMouseEvent', { 'type': 'mouseReleased', 'button': 'left', 'clickCount': 1, 'x': self._cx, 'y': self._cy})
 
     def send_keys(self, value, click_first=True):
