@@ -8,31 +8,25 @@ import requests
 import sys
 import time
 import traceback
-import urllib.parse
 import uuid
 import warnings
-import webbrowser
-
 
 import io
 from distutils.util import strtobool
 from PIL import Image
 from packaging import version
 import selenium
+from selenium.webdriver.common.action_chains import ActionChains
 
 if version.parse(selenium.__version__) < version.parse('4.0.0'):
     old_selenium = True
 else:
     old_selenium = False
 
-from selenium.webdriver.common.action_chains import ActionChains
-from selenium.common.exceptions import StaleElementReferenceException
 from selenium.webdriver.common.by import By
 
-
-from selenium import webdriver
-
 requests.packages.urllib3.disable_warnings()
+from .scanner import Scanner
 
 
 log = logging.getLogger(__name__)
@@ -70,6 +64,9 @@ class SeleniumDriverCore(object):
         self.url = initialization_options.get('server_url',
                 os.environ.get('DEVTOOLSAI_URL', 'https://smartdriver.dev-tools.ai'))
         self.default_prod_url = 'https://smartdriver.dev-tools.ai'
+
+        self.scanner = Scanner(self, self.driver, self.url, self.api_key)
+
         test_case_name = initialization_options.get('test_case_name', None)
         if test_case_name is None:
             test_case_name = traceback.format_stack()[0].split()[1].split('/')[-1].split('.py')[0]
@@ -127,6 +124,9 @@ class SeleniumDriverCore(object):
     def implicitly_wait(self, wait_time):
         self.driver.implicitly_wait(wait_time)
 
+    def scan_domain(self, domain, max_depth=10):
+        self.scanner.crawl_domain(domain, max_depth=max_depth)
+
     def find_element(self, by='id', value=None, element_name=None):
         """
         Find an element given a By strategy and locator.
@@ -142,7 +142,7 @@ class SeleniumDriverCore(object):
         #        If succesful, return element
         #        If NOT succesful, raise element not found with link
         if element_name is None:
-            element_name = 'element_name_by_%s_%s' % (str(by).replace('.', '_'), str(value).replace('.', '_'))
+            element_name = 'element_name_by_locator_By_%s:_%s' % (str(by).replace('.', '_').replace(' ', '_'), str(value).replace('.', '_').replace(' ', '_'))
         return self._generic_find_method(
             self.driver.find_element, element_name, by, value)
 
@@ -164,10 +164,7 @@ class SeleniumDriverCore(object):
             ::
                 element = driver.find_element_by_accessibility_id('foo')
         """
-        if element_name is None:
-            element_name = 'element_name_by_accessibility_id_%s' % (str(accessibility_id).replace('.', '_'))
-        return self._generic_find_method(
-            self.driver.find_element_by_accessibility_id, element_name, accessibility_id)
+        return self.find_element(self, by=By.ACCESSIBILITY_ID, value=accessibility_id, element_name=element_name)
 
     def find_element_by_class_name(self, name, element_name=None):
         """
@@ -187,10 +184,7 @@ class SeleniumDriverCore(object):
             ::
                 element = driver.find_element_by_class_name('foo')
         """
-        if element_name is None:
-            element_name = 'element_name_by_class_name_%s' % (str(name).replace('.', '_'))
-        return self._generic_find_method(
-            self.driver.find_element_by_class_name, element_name, name)
+        return self.find_element(by=By.CLASS_NAME, value=name, element_name=element_name)
 
     def find_element_by_css_selector(self, css_selector, element_name=None):
         """
@@ -211,10 +205,7 @@ class SeleniumDriverCore(object):
 
                 element = driver.find_element_by_css_selector('#foo')
         """
-        if element_name is None:
-            element_name = 'element_name_by_css_selector_%s' % (str(css_selector).replace('.', '_'))
-        return self._generic_find_method(
-            self.driver.find_element_by_css_selector, element_name, css_selector)
+        return self.find_element(by=By.CSS_SELECTOR, value=css_selector, element_name=element_name)
 
     def find_element_by_id(self, id_, element_name=None):
         """
@@ -235,10 +226,7 @@ class SeleniumDriverCore(object):
 
                 element = driver.find_element_by_id('foo')
         """
-        if element_name is None:
-            element_name = 'element_name_by_id_%s' % (str(id_).replace('.', '_'))
-        return self._generic_find_method(
-            self.driver.find_element_by_id, element_name, id_)
+        return self.find_element(by=By.ID, value=id_, element_name=element_name)
 
     def find_element_by_link_text(self, link_text, element_name=None):
         """
@@ -259,10 +247,7 @@ class SeleniumDriverCore(object):
 
                 element = driver.find_element_by_link_text('Sign In')
         """
-        if element_name is None:
-            element_name = 'element_name_by_link_text_%s' % (str(link_text).replace('.', '_'))
-        return self._generic_find_method(
-            self.driver.find_element_by_link_text, element_name, link_text)
+        return self.find_element(by=By.LINK_TEXT, value=link_text, element_name=element_name)
 
     def find_element_by_name(self, name, element_name=None):
         """
@@ -283,10 +268,7 @@ class SeleniumDriverCore(object):
 
                 element = driver.find_element_by_name('foo')
         """
-        if element_name is None:
-            element_name = 'element_name_by_name_%s' % (str(name).replace('.', '_'))
-        return self._generic_find_method(
-            self.driver.find_element_by_name, element_name, name)
+        return self.find_element(by=By.NAME, value=name, element_name=element_name)
 
     def find_element_by_partial_link_text(self, link_text, element_name=None):
         """
@@ -307,10 +289,7 @@ class SeleniumDriverCore(object):
 
                 element = driver.find_element_by_partial_link_text('Sign')
         """
-        if element_name is None:
-            element_name = 'element_name_by_partial_link_text_%s' % (str(link_text).replace('.', '_'))
-        return self._generic_find_method(
-            self.driver.find_element_by_partial_link_text, element_name, link_text)
+        return self.find_element(by=By.PARTIAL_LINK_TEXT, value=link_text, element_name=element_name)
 
     def find_element_by_tag_name(self, name, element_name=None):
         """
@@ -331,10 +310,7 @@ class SeleniumDriverCore(object):
 
                 element = driver.find_element_by_tag_name('h1')
         """
-        if element_name is None:
-            element_name = 'element_name_by_tag_name_%s' % (str(name).replace('.', '_'))
-        return self._generic_find_method(
-            self.driver.find_element_by_tag_name, element_name, name)
+        return self.find_element(by=By.TAG_NAME, value=name, element_name=element_name)
 
     def find_element_by_xpath(self, xpath, element_name=None):
         """
@@ -355,10 +331,7 @@ class SeleniumDriverCore(object):
 
                 element = driver.find_element_by_xpath('//div/td[1]')
         """
-        if element_name is None:
-            element_name = 'element_name_by_xpath_%s' % (str(xpath).replace('.', '_'))
-        return self._generic_find_method(
-            self.driver.find_element_by_xpath, element_name, xpath)
+        return self.find_element(by=By.XPATH, value=xpath, element_name=element_name)
 
     def find_by_ai(self, element_name):
         """
@@ -394,11 +367,11 @@ class SeleniumDriverCore(object):
             driver_element = find_method(*args)
             if driver_element:
                 try:
-                    key = self._upload_screenshot_if_necessary(element_name)
+                    key = self._upload_screenshot_if_necessary(element_name, driver_element=driver_element)
                     if key is not None:
                         # Key is None when element is frozen or another issue during screenshot and upload.
                         self._update_elem(driver_element, key, element_name)
-                except Exception:
+                except Exception as e:
                     log.error('error uploading screenshot to devtools. Continuing.')
             return driver_element
         except NoElementFoundException as e:
@@ -441,13 +414,20 @@ class SeleniumDriverCore(object):
             'width': elem.rect['width'] * self.multiplier,
             'height': elem.rect['height'] * self.multiplier,
             'multiplier': self.multiplier,
-            'test_case_name': self.test_case_uuid
+            'test_case_name': self.test_case_uuid,
+            'page_offset': self.page_offset * self.multiplier,
+            'ref_screenshot_uuid': self.ref_screenshot_uuid
         }
         try:
             action_url = self.url + '/add_action_info'
             # Verify is False as the lets encrypt certificate raises issue on mac.
-            res = requests.post(action_url, json=data, verify=False)
-            res = res.json()
+            for _ in range(2):
+                try:
+                    res = requests.post(action_url, json=data, verify=False, timeout=4)
+                    res = res.json()
+                    break
+                except Exception:
+                    log.error('error loading add_action_info')
             if not res['success']:
                 log.error(res['message'])
         except Exception:
@@ -472,7 +452,12 @@ class SeleniumDriverCore(object):
         data = {'api_key': self.api_key, 'screenshot_uuid': screenshot_uuid, 'label': element_name}
         check_screenshot_url = self.url + '/exists_screenshot'
         start = time.time()
-        r = requests.post(check_screenshot_url, json=data, verify=False).json()
+        for _ in range(2):
+            try:
+                r = requests.post(check_screenshot_url, json=data, verify=False, timeout=6).json()
+                break
+            except Exception:
+                log.error('error in exists_screenshot')
         end = time.time()
         if self.debug:
             print(f'Cached bounding box request time: {end - start}')
@@ -486,16 +471,38 @@ class SeleniumDriverCore(object):
         request_data = {'api_key': self.api_key,
                         'label': element_name
                         }
-        resp = requests.post(url, json=request_data).json()
+        for _ in range(2):
+            try:
+                resp = requests.post(url, json=request_data, timeout=3).json()
+                break
+            except Exception:
+                log.error('error in check_frozen')
         return resp.get('is_frozen', True)
 
 
-    def _upload_screenshot_if_necessary(self, element_name):
+    def _upload_screenshot_if_necessary(self, element_name, driver_element=None):
         if self._check_frozen(element_name):
             return None
 
         screenshotBase64 = self._get_screenshot()
         screenshot_uuid = self.get_screenshot_hash(screenshotBase64)
+        self.ref_screenshot_uuid = None
+        self.page_offset = 0
+        if driver_element is not None:
+            self.page_offset = self.driver.execute_script('return window.pageYOffset')
+            needs_to_scroll = (driver_element.rect['y'] > (self.window_size['height'] + self.page_offset)) or (driver_element.rect['y'] < self.page_offset)
+            if needs_to_scroll:
+                self.previous_page_offset = self.page_offset
+                self.ref_screenshot_uuid = screenshot_uuid
+                ActionChains(self.driver) \
+                    .scroll_to_element(driver_element) \
+                    .perform()
+                screenshotBase64 = self._get_screenshot()
+                screenshot_uuid = self.get_screenshot_hash(screenshotBase64)
+                self.page_offset = self.driver.execute_script('return window.pageYOffset')
+                ActionChains(self.driver).scroll_by_amount(0, int(self.previous_page_offset - self.page_offset)).perform()
+                self.page_offset = self.driver.execute_script('return window.pageYOffset')
+
         # Check results
         try:
             response = self._check_screenshot_exists(screenshot_uuid, element_name)
@@ -518,7 +525,12 @@ class SeleniumDriverCore(object):
                         'test_case_name': self.test_case_uuid}
                 upload_screenshot_url = self.url + '/upload_screenshot'
                 start = time.time()
-                r = requests.post(upload_screenshot_url, json=data, verify=False).json()
+                for _ in range(2):
+                    try:
+                        r = requests.post(upload_screenshot_url, json=data, verify=False, timeout=6).json()
+                        break
+                    except Exception:
+                        log.error('error uploading url')
                 end = time.time()
                 if r['success'] == False:
                     log.error(f'Error uploading screenshot {screenshot_uuid}')
@@ -531,20 +543,30 @@ class SeleniumDriverCore(object):
         except Exception:
             log.exception('Error checking cached screenshot / uploading it from remote')
 
-    def _test_case_get_box(self, label):
+    def _test_case_get_box(self, label, event_id=None):
         """
             Checks for a bounding box given the last screenshot uuid that we got when uploading it.
         """
         box = None
-        data = {'api_key': self.api_key, 'label': label, 'screenshot_uuid': self.last_test_case_screenshot_uuid,
-                'run_classifier': self.use_classifier_during_creation}
+        data = {'api_key': self.api_key, 'label': label,
+                'screenshot_uuid': self.last_test_case_screenshot_uuid,
+                'run_classifier': self.use_classifier_during_creation,
+                'event_id': event_id}
 
-        r = requests.post(self.url + '/testcase/get_action_info', json=data, verify=False).json()
+        needs_reload = False
+        r = {'success': False}
+        for _ in range(2):
+            try:
+                r = requests.post(self.url + '/testcase/get_action_info', json=data, verify=False, timeout=3).json()
+                break
+            except Exception:
+                log.error('error getting test case actioninfo')
         if r['success']:
             box = r['predicted_element']
+            needs_reload = r['needs_reload']
         else:
             log.error(r['message'])
-        return box
+        return box, needs_reload
 
 
     def _test_case_upload_screenshot(self, label):
@@ -559,7 +581,12 @@ class SeleniumDriverCore(object):
                 'screenshot': screenshotBase64,
                 'label': label,
                 'is_interactive': True}
-        res = requests.post(url, json=data, verify=False).json()
+        for _ in range(2):
+            try:
+                res = requests.post(url, json=data, verify=False, timeout=6).json()
+                break
+            except Exception:
+                log.error('error uploading screenshot')
         if res['success']:
             self.last_test_case_screenshot_uuid = res['screenshot_uuid']
             self.last_screenshot = screenshotBase64
